/****************************************************************************
    Xplanet 0.90 - render an image of a planet into an X window
    Copyright (C) 2001 Hari Nair <hari@alumni.caltech.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
****************************************************************************/

#include <cmath>
#include <cstdio>
#include <fstream>
#include <iostream>
#include <string>
using namespace std;

#include <unistd.h>

#include "auxfiles.h"
#include "Location.h"
#include "Options.h"
#include "parse.h"
#include "Satellite.h"
#include "util.h"

static list<Satellite> satellite_list;

static bool
satelliteFileExists(string &filename)
{
    if (opts.debug) cout << "\nLooking for satellite file ...\n";

    if (opts.debug) cout << filename << endl;
    if (fopen(filename.c_str(), "r") != NULL) return(true);

    string satellite_file = "satellites";
    satellite_file += separator;
    satellite_file += filename;
    
    if (opts.debug) cout << satellite_file << endl;
    if (fopen(satellite_file.c_str(), "r") != NULL)
    {
        filename = satellite_file;
        return(true);
    }

    satellite_file = prefix;
    satellite_file += separator;
    satellite_file += "satellites";
    satellite_file += separator;
    satellite_file += filename;

    if (opts.debug) cout << satellite_file << endl;
    if (fopen(satellite_file.c_str(), "r") != NULL)
    {
        filename = satellite_file;
        return(true);
    }

    return(false);
}

/*
  Given an value X and radius d, this routine draws a quarter-circle
  in the plane where X is constant.  
*/
static void
drawAltitudeQuad(const double rot[3][3], const double X, const double d,
		 bool Y_positive, bool Z_positive, ofstream &sat_a, 
		 const string color, const double spacing)
{
    double Y = d;
    double Z = (1 - X*X - Y*Y);
    if (fabs(Z) < 1e-5) Z = 0;
    Z = sqrt(Z);
    if (!Y_positive) Y *= -1;
    if (!Z_positive) Z *= -1;

    Location point(X, Y, Z);
    point.rotate(rot);
    point.rectangularToSpherical();

    for (double siny = 1; siny >= 0; siny -= 1./10.)
    {
	double prev_lat = point.lat;
	double prev_lon = point.lon;
	
	Y = sin(M_PI_2 * siny) * d;
	Z = (1 - X*X - Y*Y);
	if (fabs(Z) < 1e-5) Z = 0;
	Z = sqrt(Z);
	if (!Y_positive) Y *= -1;
	if (!Z_positive) Z *= -1;

	point.setLocation(X, Y, Z);
	point.rotate(rot);
	point.rectangularToSpherical();
	sat_a << prev_lat/deg_to_rad << '\t' << prev_lon/deg_to_rad 
	      << '\t' << point.lat/deg_to_rad 
	      << '\t' << point.lon/deg_to_rad;
	if (!color.empty()) sat_a << " color=" << color;
	if (spacing > 0) sat_a << " spacing=" << spacing;

	sat_a << endl;
    }
}

static void
readSatFileLine(const char *line, ofstream &sat_m, ofstream &sat_a)
{
    int i = 0;
    while (isDelimiter(line[i])) i++;
    if (isEndOfLine(line[i])) return;

    int id;
    sscanf(line + i, "%d", &id);
    while (!(isDelimiter(line[i]) || isEndOfLine(line[i]))) i++; 

    Satellite *this_sat = NULL;
    list<Satellite>::iterator p = satellite_list.begin();
    while (p != satellite_list.end())
    {
	if (id == p->getID())
	{
	    this_sat = &(*p);
	    if (opts.debug) 
	    {
		cout << "Found ID #" << id << " (" << this_sat->getName() 
		     << ")\n";
	    }
	    break;
	}
	p++;
    }

    if (this_sat == NULL) 
    {
	cerr << "Satellite ID #" << id << " not found in TLE file\n";
	return;
    }

    int align = AUTO;
    list<double> altcirc;
    string color;
    
    string font;
    int fontsize = opts.fontsize;
    
    string image;
    string name = this_sat->getName();
    
    char *returnstring = NULL;
    
    float spacing = -1;

    int trail_type = ORBIT;
    int trail_length = 0;
    int trail_interval = 1;
    string transparency;
    
    bool end_of_line = false;
    bool syntax_error = false;

    while (!end_of_line)
    {
        switch (parse(i, line, returnstring))
        {
        case ALIGN:
            if (returnstring == NULL) break;
            switch (returnstring[0])
            {
            case 'r':
            case 'R':
                align = RIGHT;
                break;
            case 'l':
            case 'L':
                align = LEFT;
                break;
            case 'a':
            case 'A':
                align = ABOVE;
                break;
            case 'b':
            case 'B':
                align = BELOW;
                break;
            case 'c':
            case 'C':
                align = CENTER;
                break;
            default:
                cerr << "Unrecognized option for align in "
                     << "satellite file" << endl;
                break;
            }
            break;
	case ALTCIRC:
	{
	    double angle;
            if (returnstring == NULL) break;
            sscanf(returnstring, "%lf", &angle);
	    if (angle < 0) angle *= -1;
	    if (angle > 90) angle = 90;
	    angle += 90;
	    altcirc.push_back(angle * deg_to_rad);
	}
	break;
        case COLOR:
            if (returnstring == NULL) break;
            color.assign(returnstring);
            break;
        case DELIMITER:
            break;
        case ENDOFLINE:
            end_of_line = true;
            break;
        case FONT:
            if (returnstring == NULL) break;
            font.assign(returnstring);
            break;
        case FONTSIZE:
            if (returnstring == NULL) break;
            sscanf(returnstring, "%d", &fontsize);
            if (fontsize < 0) fontsize = opts.fontsize;
            break;
        case IMAGE:
            if (returnstring == NULL) break;
            image.assign(returnstring);
            break;
        case NAME:
            if (returnstring == NULL) break;
            name.assign(returnstring);
            break;
        case SPACING:
	    if (returnstring == NULL) break;
	    sscanf(returnstring, "%f", &spacing);
	    if (spacing < 0.01) spacing = 0.01;
	    break;
	case TRAIL:
	{
	    char *ptr = returnstring;
	    while (ptr[0] != ',') 
	    {
		if (ptr[0] == '\0') 
		{
		    syntax_error = true;
		    break;
		}
		ptr++;
	    }

	    if (syntax_error) break;

            if (!sscanf(++ptr, "%d,%d", &trail_length, &trail_interval) == 2)
	    {
		cerr << "Need three values for trail{}!\n";
		syntax_error = true;
	    }
	    else
	    {
		switch (returnstring[0])
		{
		case 'g':
		case 'G':
		    trail_type = GROUND;
		    break;
		case 'o':
		case 'O':
		    trail_type = ORBIT;
		    break;
		default:
		    cerr << "Unknown type of orbit trail!\n";
		    syntax_error = true;
		    break;
		}
		if (trail_length < 0) trail_length = 0;
		if (trail_interval < 1) trail_interval = 1;
	    }
	}
	break;
        case TRANSPARENT:
            if (returnstring == NULL) break;

            transparency.assign(returnstring);
	    
            int r, g, b;
            if (!sscanf(returnstring, "%d,%d,%d", &r, &g, &b) == 3)
            {
                cerr << "Need three values for transparency pixel!\n";
                syntax_error = true;
            }

            break;
        default:
        case UNKNOWN:
            syntax_error = true;
            break;
        }

        if (returnstring != NULL)
        {
            delete [] returnstring;
            returnstring = NULL;
        }

        if (syntax_error)
        {
            cerr << "Syntax error in satellite file\n";
            cerr << "line is \"" << line << "\"" << endl;
            return;
        }

        i++;
    }

    const double EARTH_RADIUS = 6378.140;

    // Reload TLE data here since select_ephemeris() in libsgp4sdp4
    // changes it.  This is in case the user wants to have two entries
    // with the same satellite.
    this_sat->loadTLE();
    
    double lat = 0, lon = 0, alt = 0;
    
    time_t end_time = (time_t) opts.time.tv_sec;
    time_t start_time = end_time - trail_length * 60;
    time_t interval = trail_interval * 60;
    
    this_sat->getSpherical(start_time, lat, lon, alt);
    
    for (time_t t = start_time + interval; t <= end_time; t += interval)
    {
	double prev_lat = lat;
	double prev_lon = lon;
	double prev_alt = alt;
	this_sat->getSpherical(t, lat, lon, alt);
	sat_a << prev_lat/deg_to_rad << '\t' << prev_lon/deg_to_rad 
	      << '\t' << lat/deg_to_rad << '\t' << lon/deg_to_rad;
	
	if (!color.empty()) sat_a << " color=" << color;
	if (spacing > 0) sat_a << " spacing=" << spacing;
	if (trail_type == ORBIT)
	    sat_a << " radius=" << 1 + prev_alt/EARTH_RADIUS
		  << " radius=" << 1 + alt/EARTH_RADIUS;
	sat_a << endl;
    }
    
    this_sat->getSpherical((time_t) opts.time.tv_sec, lat, lon, alt);
    
    sat_m << lat/deg_to_rad << ' ' << lon/deg_to_rad;
    if (trail_type == ORBIT)
	sat_m << " radius=" << 1 + alt/EARTH_RADIUS;
    if (align != AUTO) 
    {
	sat_m << " align=";
	switch (align)
	{
	case RIGHT:
	    sat_m << "right";
	    break;
	case ABOVE:
	    sat_m << "above";
	    break;
	case BELOW:
	    sat_m << "below";
	    break;
	case LEFT:
	    sat_m << "left";
	    break;
	case CENTER:
	    sat_m << "center";
	    break;
	default:
	    cerr << "Unknown alignment????\n";
	}
    }
    if (!color.empty()) sat_m << " color=" << color;
    if (!font.empty()) sat_m << " font=" << font;
    if (fontsize != opts.fontsize) sat_m << " fontsize=" << fontsize;
    if (!image.empty()) sat_m << " image=" << image;
    if (!name.empty()) sat_m << " {" << name << "}";
    if (!transparency.empty()) 
	sat_m << " transparent={" << transparency << "}";

    sat_m << endl;
    
    // Distance of satellite from Earth's center, in units of Earth radii
    double Rs = 1 + alt/EARTH_RADIUS;
    
    double rot[3][3];
    rotateXYZ(rot, 0, lat, -lon);

    list<double>::iterator a = altcirc.begin();
    while (a != altcirc.end())
    {
	double phi = asin(sin(*a) / Rs);
	double X = -cos(*a + phi);
	double d = sin(*a + phi);

	drawAltitudeQuad(rot, X, d, true, true, sat_a, color, spacing);
	drawAltitudeQuad(rot, X, d, true, false, sat_a, color, spacing);
	drawAltitudeQuad(rot, X, d, false, true, sat_a, color, spacing);
	drawAltitudeQuad(rot, X, d, false, false, sat_a, color, spacing);

	a++;
    }
}

void
createSatelliteList()
{
    list<string> remove_list;

    list<string>::iterator p = opts.satellitefile.begin();
    while(p != opts.satellitefile.end())
    {
	char lines[3][80];
	if (satelliteFileExists(*p))
	{
	    if (opts.debug) cout << "Opening satellite file " << *p << endl;
	
	    string tlefile = *p + ".tle";
	    ifstream infile(tlefile.c_str());
	    if (infile.bad())
	    {
		cerr << "Can't open TLE file " << tlefile << endl;
		remove_list.push_back(*p);
	    }

	    while (infile.getline(lines[0], 80) != NULL)
	    {
		if ((infile.getline(lines[1], 80) == NULL) 
		    || (infile.getline(lines[2], 80) == NULL))
		{
		    cerr << "Malformed TLE file?\n";
		    break;
		}
		
		Satellite sat(lines);
		
		if (!sat.isGoodData()) 
		{
		    cerr << "Bad TLE data\n";
		    continue;
		}
		if (opts.debug) cout << "Read satellite with ID "
				     << sat.getID() << endl;

		satellite_list.push_back(sat);
	    }

	    infile.close();
	}
	else
	{
	    cerr << "Can't open satellite file " << *p << endl;
	    remove_list.push_back(*p);
	}
	p++;
    }
    
    p = remove_list.begin();
    while (p != remove_list.end())
    {
	opts.satellitefile.remove(*p);
	p++;
    }
}

void
createSatelliteFiles(string tmpdir)
{
    int count = 0;
    list<string>::iterator p = opts.satellitefile.begin();
    while(p != opts.satellitefile.end())
    {
	if (count > 99)
	{
	    cerr << "Sorry, only 100 satellite files max!\n";
	    break;
	}
	char pid[9];
	sprintf(pid, "%5.5d-%2.2d", ((int) (getpid() % 10000)), count++);
	
	string sat_markers = tmpdir + separator + "xplanet-markers." + pid;
	string sat_arcs = tmpdir + separator + "xplanet-arcs." + pid;
	
	ofstream sat_m(sat_markers.c_str());
	ofstream sat_a(sat_arcs.c_str());

	if (opts.debug) cout << "Opening satellite file " << *p << endl;
	ifstream infile(p->c_str());
	
	char *line = new char[256];
	while (infile.getline (line, 256, '\n') != NULL)
	    readSatFileLine(line, sat_m, sat_a);
	
	infile.close();

	sat_m.close();
	sat_a.close();
	
	opts.markerfile.push_back(sat_markers);
	opts.greatarcfile.push_back(sat_arcs);
	
	p++;
    }
}

void
cleanupSatelliteFiles(string tmpdir)
{
    int count = 0;
    list<string>::iterator p = opts.satellitefile.begin();
    while(p != opts.satellitefile.end())
    {
	if (count > 99) break;

	char pid[9];
	sprintf(pid, "%5.5d-%2.2d", ((int) (getpid() % 10000)), count++);
	
	string sat_markers = tmpdir + separator + "xplanet-markers." + pid;
	string sat_arcs = tmpdir + separator + "xplanet-arcs." + pid;

	unlink(sat_markers.c_str());
	unlink(sat_arcs.c_str());
	p++;
    }
}

bool
getSatLocation(const int id, const time_t time, double &lat, double &lon,
	       double &alt)
{
    Satellite *this_sat = NULL;
    list<Satellite>::iterator p = satellite_list.begin();
    while (p != satellite_list.end())
    {
	if (id == p->getID())
	{
	    this_sat = &(*p);
	    if (opts.debug) 
	    {
		cout << "Found ID #" << id << " (" << this_sat->getName() 
		     << ")\n";
	    }
	    break;
	}
	p++;
    }

    if (this_sat == NULL) 
    {
	if (satellite_list.size() == 0) 
	{
	    cerr << "A TLE file must be specified using the "
		 << "-satfile option.\n";
	}
	else
	{
	    cerr << "Satellite ID #" << id << " not found in TLE file\n";
	}
	return(false);
    }

    // Reload TLE data here since select_ephemeris() in libsgp4sdp4
    // changes it.  This is in case the user wants to have two entries
    // with the same satellite.
    this_sat->loadTLE();
    
    this_sat->getSpherical(time, lat, lon, alt);
    return(true);
}

