/****************************************************************************
    Xplanet 0.90 - render an image of a planet into an X window
    Copyright (C) 2001 Hari Nair <hari@alumni.caltech.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
****************************************************************************/

#include <string>
using namespace std;

#include "DisplayMagickpp.h"
#include "Options.h"
#include "util.h"

#include "libannotate/libannotate.h"
 
using namespace Magick;

DisplayMagickpp::DisplayMagickpp()
{
    window_width = opts.window_width;
    window_height = opts.window_height;

    // (x,y) of the globe's center
    if (opts.centx < 0 || opts.centy < 0)
    {
        opts.centx = window_width / 2;
        opts.centy = window_height / 2;
    }
  
    window_area = window_width * window_height;
}

void
DisplayMagickpp::initialize()
{
    if (opts.label || opts.markers)
    {
        fg.assign("White");
        bg.assign("Black");

        loadRGB();
    }
}

const char *
DisplayMagickpp::identifyMe()
{
    return("Magick++");
}

void 
DisplayMagickpp::render(ProjectionBase *image)
{
    thisImage = new Image(window_width, window_height, "RGB", 
                          CharPixel, image->rgb_data);
    thisImage->strokeWidth(1);

    if (!opts.font.empty()) currentfont = opts.font;
    if (opts.label || opts.markers)
    {
        setColor(opts.color);
        setFont(opts.font);
        if (opts.markers) drawMarkers(image, this);
        if (opts.label) createLabel();
    }
    drawSatelliteTracks(image, this);

    try
    {
        if (opts.display == WINDOW)
            thisImage->display();
        else
            thisImage->write(opts.output);
    }
    catch (Exception e) 
    { 
        cerr << "Caught exception while writing file: " << e.what() << endl; 
    } 

    return;
}

int
DisplayMagickpp::getFontAscent()
{
    return(font_ascent);
}

int
DisplayMagickpp::getFontDescent()
{
    return(font_descent);
}

int
DisplayMagickpp::getTextHeight()
{
    return(font_ascent + font_descent);
}

int
DisplayMagickpp::getTextHeight(string text)
{
    return(getTextHeight());
}

int
DisplayMagickpp::getTextWidth(string text)
{
    getFontMetrics(currentfont, text, string_width, 
                   font_ascent, font_descent);
    return(string_width);
}

void 
DisplayMagickpp::setColor(string color)
{
    string defaultcolor = "red";
    
    if (color.empty())
        fg = defaultcolor;
    else
        fg.assign(color);

    try
    {
        Image test;
        test.strokeColor(fg);
    }
    catch (Exception e)
    {
        cerr << "Can't find color " << color << ", using " << defaultcolor
             << endl;
        fg = defaultcolor;
    }
}

void 
DisplayMagickpp::setFont(string fontname)
{
    string defaultFont = "Helvetica";
    string thisFont;

    if (fontname.empty())
        thisFont = defaultFont;
    else
        thisFont.assign(fontname);

    try
    {
        thisImage->font(thisFont);
    }
    catch (Exception e)
    {
        cerr << "Can't load font " << fontname << ", using " << defaultFont 
             << endl;
        thisFont = defaultFont;
        thisImage->font(thisFont);
    }
    currentfont = thisFont;
    thisImage->fontPointsize(12);
}

void 
DisplayMagickpp::drawCircle(const int x, const int y)
{
    Color savePixel = thisImage->pixelColor(x, y);
    thisImage->penColor(bg);
    thisImage->draw(DrawableCircle(x, y, x-3, y));

    thisImage->penColor(fg);
    thisImage->draw(DrawableCircle(x, y, x-2, y));

    thisImage->penColor(bg);
    thisImage->draw(DrawableCircle(x, y, x-1, y));
    thisImage->penColor(fg);

    thisImage->pixelColor(x, y, savePixel);
}

void 
DisplayMagickpp::drawLabel(const string body_label, const string date_label, 
			   const string illum_label, const string obs_label, 
			   const string sun_label)
{
    if (opts.label_mask & XNegative) opts.label_x += window_width;

    if (opts.label_mask & YNegative) opts.label_y += window_height;

    int x = opts.label_x;
    int y = opts.label_y;

    getFontMetrics(currentfont, body_label, string_width, string_height);

    if (opts.label_mask & YNegative)
    {
        y = opts.label_y - 3 * string_height;
	if (opts.label_name) y -= string_height;
        if (opts.projection == ORTHOGRAPHIC) y -= string_height;
    }

    if (opts.label_name)
    {
	if (opts.label_mask & XNegative) 
	    x = opts.label_x - string_width;
	y += string_height;
	drawOutlinedString(x, y, body_label);
    }

    getFontMetrics(currentfont, date_label, string_width, string_height);
    if (opts.label_mask & XNegative) 
        x = opts.label_x - string_width;
    y += string_height;
    drawOutlinedString(x, y, date_label);

    getFontMetrics(currentfont, sun_label, string_width, string_height);
    if (opts.label_mask & XNegative) 
        x = opts.label_x - string_width;
    y += string_height;
    drawOutlinedString(x, y, sun_label);

    getFontMetrics(currentfont, obs_label, string_width, string_height);
    if (opts.label_mask & XNegative) 
        x = opts.label_x - string_width;
    y += string_height;
    drawOutlinedString(x, y, obs_label);
  
    if (opts.projection == ORTHOGRAPHIC)
    {
        float phase;
        char illum2[64];

        sscanf(illum_label.c_str(), "illumination %f", &phase);
        sprintf(illum2, "illumination %.3f", phase / 100);
	
        getFontMetrics(currentfont, illum2, string_width, string_height);
        if (opts.label_mask & XNegative) 
            x = opts.label_x - string_width;
        y += string_height;
        drawOutlinedString(x, y, illum2);
    }
}

void
DisplayMagickpp::drawOutlinedString(const int x, const int y, 
                                    const string text)
{
    thisImage->penColor(bg);
    thisImage->annotate(string(text), Geometry(0, 0, x+1, y));
    thisImage->annotate(string(text), Geometry(0, 0, x-1, y));
    thisImage->annotate(string(text), Geometry(0, 0, x, y+1));
    thisImage->annotate(string(text), Geometry(0, 0, x, y-1));
    thisImage->penColor(fg);
    thisImage->annotate(string(text), Geometry(0, 0, x, y));
}
