/****************************************************************************
    Xplanet 0.90 - render an image of a planet into an X window
    Copyright (C) 2001 Hari Nair <hari@alumni.caltech.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
****************************************************************************/

#include <climits>
#include <cstdlib>
#include <fstream>
#include <iostream>
using namespace std;

#include <X11/keysym.h>
#include <X11/Xatom.h>

#include "auxfiles.h"
#include "DisplayX11.h"
#include "Marker.h"
#include "Options.h"
#include "util.h"

#include "libannotate/libannotate.h"
#include "libimage/libimage.h"

DisplayX11::DisplayX11() : DisplayX11Base() 
{
}

void
DisplayX11::initialize()
{
    if (opts.label || opts.markers || !opts.satellitefile.empty())
    {
	bg_pixel       = BlackPixel(display, DefaultScreen(display));

        setColor(opts.color);
        setFont(opts.font);
    }
}

int
DisplayX11::getFontAscent()
{
    return(font->ascent);
}

int
DisplayX11::getFontDescent()
{
    return(font->descent);
}

void
DisplayX11::getRGB(string color, unsigned char *RGB)
{
    string defaultcolor = "red";
    if (color.empty()) color = defaultcolor;

    XColor xc;
    if (!XParseColor(display, colormap, color.c_str(), &xc))
    {
        cerr << "Can't find color " << color << ", using " << defaultcolor
             << endl;
        XParseColor(display, colormap, defaultcolor.c_str(), &xc); 
    }

    RGB[0] = (unsigned char) (xc.red >> 8);
    RGB[1] = (unsigned char) (xc.green >> 8);
    RGB[2] = (unsigned char) (xc.blue >> 8);
}

int
DisplayX11::getTextHeight()
{
    return(font->ascent + font->descent);
}

int
DisplayX11::getTextHeight(string text)
{
    return(getTextHeight());
}

int
DisplayX11::getTextWidth(string text)
{
    return(XTextWidth(font, text.c_str(), text.size()));
}

void
DisplayX11::setColor(string color)
{
    string defaultcolor = "red";
    if (color.empty()) color = defaultcolor;
  
    XColor xc;      
    if (!XAllocNamedColor(display, colormap, color.c_str(), &xc, &xc)) 
    {
        cerr << "Can't find color " << color << ", using " << defaultcolor
             << endl;
        XAllocNamedColor(display, colormap, defaultcolor.c_str(), &xc, &xc);
    }
 
    fg_pixel = xc.pixel;
}

void
DisplayX11::setFontSize(const int size)
{
// overrides DisplayBase::setFontSize(const int size)
}

void
DisplayX11::setFont(string fontname)
{
    string defaultfont = "variable";
    if (fontname.empty()) fontname = defaultfont;
    font = XLoadQueryFont(display, fontname.c_str());

    if (font == NULL)
    {
        cerr << "Can't load font " << fontname << ", using " << defaultfont 
             << endl;
        font = XLoadQueryFont(display, defaultfont.c_str());
    }
  
    XSetFont(display, gc, font->fid);
}

void
DisplayX11::drawImage(Marker &m)
{
    if (m.image_width == 0) return;

    if (m.image_data == NULL)
    {
	drawCircle(m.x, m.y);
	return;
    }

    XImage *ximage = XGetImage(display, pixmap, 0, 0, window_width, 
                               window_height, AllPlanes, ZPixmap);

    Pixmap new_pixmap = createPixmap(m.image_data, m.image_width, 
				     m.image_height);

    XImage *new_image = XGetImage(display, new_pixmap, 0, 0, m.image_width, 
				  m.image_height, AllPlanes, ZPixmap);

    int offset_x = m.x - m.image_width/2;
    int offset_y = m.y - m.image_height/2;
    for (int j = 0; j < m.image_height; j++)
    {
	int pixel_y = j + offset_y;
	if (pixel_y < 0 || pixel_y >= window_height) continue;
	for (int i = 0; i < m.image_width; i++)
	{
	    int pixel_x = i + offset_x;
	    if (pixel_y == window_height - 1 && pixel_x >= window_width)
		pixel_x -= window_width;

	    if (!m.isTransparent(i, j))
	    {
		XPutPixel(ximage, pixel_x, pixel_y, 
			  XGetPixel(new_image, i, j));
	    }
	}
    }
    
    XPutImage(display, pixmap, gc, ximage, 0, 0, 0, 0, window_width, 
	      window_height);
}

void
DisplayX11::drawCircle(const int x, const int y)
{
    // the last argument to XDrawArc is the angle in 64ths of a degree
    XSetForeground(display, gc, bg_pixel);
    XDrawArc(display, pixmap, gc, x-3, y-3, 6, 6, 0, 23040);
    XDrawArc(display, pixmap, gc, x-1, y-1, 2, 2, 0, 23040);
    XSetForeground(display, gc, fg_pixel);         
    XDrawArc(display, pixmap, gc, x-2, y-2, 4, 4, 0, 23040);
}

void 
DisplayX11::drawOutlinedString(const int x, const int y, const string text)
{
    int len = text.size();
    const char *cstring = text.c_str();
    XSetForeground(display, gc, bg_pixel);
    XDrawString(display, pixmap, gc, x+1, y, cstring, len);
    XDrawString(display, pixmap, gc, x-1, y, cstring, len);
    XDrawString(display, pixmap, gc, x, y+1, cstring, len);
    XDrawString(display, pixmap, gc, x, y-1, cstring, len);
    XSetForeground(display, gc, fg_pixel);
    XDrawString(display, pixmap, gc, x, y, cstring, len);
}
