/****************************************************************************
    Xplanet 0.90 - render an image of a planet into an X window
    Copyright (C) 2001 Hari Nair <hari@alumni.caltech.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
****************************************************************************/

#include <cmath>
#include <cstdio>
#include <cstring>
#include <ctime>
#include <list>
#include <iostream>
#include <string>
using namespace std;

#include <sys/time.h>

#include "auxfiles.h"
#include "Options.h"
#include "util.h"
#include "xplanet.h"

#include "libimage/libimage.h"
#include "libprojection/libprojection.h"

#include "checkfuncs.h"

static double sunlat, sunlon, sundist;
static double moonlat, moonlon, moondist;

#ifdef HAVE_LIBMAGICK

/*
  This stuff is here instead of in the DisplayMagickpp class since it 
  depends on the ImageMagick API, not Magick++.  Evidently you can't 
  mix the two in the same source file.
*/

#include <magick/api.h>

void
getFontMetrics(const string font, const string text, 
               int &width, int &ascent, int &descent)
{
    ImageInfo image_info;
    DrawInfo draw_info;
    FontMetric metrics;

    GetImageInfo(&image_info);
    GetDrawInfo(&image_info, &draw_info);

    draw_info.text = (char *) text.c_str();
    draw_info.font = (char *) font.c_str();

    if (GetFontMetrics(NULL, &draw_info, &metrics))
    {
        width = metrics.width;
        ascent = metrics.ascent;
        descent = -metrics.descent;
    }
    else
    {
        cerr << "Can't get font metrics.\n";
    }
}

void
getFontMetrics(const string font, const string text, int &width, int &height)
{
    int ascent, descent;
    getFontMetrics(font, text, width, ascent, descent);
    height = ascent - descent;
}

#endif /* HAVE_LIBMAGICK */

double 
poly(const double a1, const double a2, const double a3, const double a4,
     const double t)
{
    return(a1 + t*(a2 + t*(a3 + t*a4)));
}

double 
julian(int year, int month, int day, int hour, int min, int sec)
{
    if(month < 3) 
    {
        year -= 1;
        month += 12;
    }      
  
    // assume it's the Gregorian calendar (after 1582 Oct 15)

    int a = year/100;
    int b = 2 - a + a/4;
    int c = int(365.25 * year);
    int d = int(30.6001 * (month + 1));
    double e = day + ((sec/60. + min) / 60. + hour) / 24.;

    double jd = b + c + d + e + 1720994.5;

    return(jd);
}

double 
julianCentury(const time_t tv_sec)
{
    double jd = julian(gmtime(&tv_sec)->tm_year + 1900,
                       gmtime(&tv_sec)->tm_mon + 1,
                       gmtime(&tv_sec)->tm_mday,
                       gmtime(&tv_sec)->tm_hour,
                       gmtime(&tv_sec)->tm_min,
                       gmtime(&tv_sec)->tm_sec);
    
    return((jd - 2415020)/36525);
}

double 
kepler(const double e, const double M)
{
    double E = M;
    double delta = 1;

    while (fabs(delta) > 1E-10)
    {
        delta = (M + e * sin(E) - E)/(1 - e * cos(E));
        E += delta;
    }
    return(E);
}

double
gmst(const double T, const time_t tv_sec)
{      
    // Sidereal time at Greenwich at 0 UT
    double g = poly(6.6460656, 2400.051262, 0.00002581, 0, T);

  // Now find current sidereal time at Greenwich
    double currgmt = (gmtime(&tv_sec)->tm_hour 
                      + gmtime(&tv_sec)->tm_min/60.
                      + gmtime(&tv_sec)->tm_sec/3600.);
    currgmt *= 1.002737908;
    g += currgmt;
    g = fmod(g, 24.0);
    if (g < 0) g += 24;
    return(g);
}

double
calcObliquity(const double T)
{
    return (poly(23.452294, -1.30125E-2, -1.64E-6, 5.03E-7, T)
            * deg_to_rad);
}

void 
calcRaDec(const double lon, const double lat, const double eps,
          double &alpha, double &delta)
{
    delta = asin(sin(eps) * sin(lon) * cos(lat) + sin(lat) * cos(eps));

    alpha = (atan2(cos(eps) * sin(lon) - tan (lat) * sin(eps), cos(lon)) 
             / deg_to_rad);
    alpha /= 15;
    alpha = fmod(alpha, 24.0);
    if (alpha < 0) alpha += 24;
}

void rotateXYZ(double matrix[3][3], const double angle_x, 
               const double angle_y, const double angle_z)
{
    if (angle_x == 0 && angle_y == 0 && angle_z == 0) 
    {
	for (int j = 0; j < 3; j++)
	    for (int i = 0; i < 3; i++)
		matrix[j][i] = (i == j ? 1 : 0 );
	return;
    }

    /*
      matrix to first rotate reference frame angle_x radians through 
      x axis, then angle_y radians through y axis, and lastly 
      angle_z radians through z axis.  Positive rotations are counter-
      clockwise looking down the axis.
    */

    double cosx = cos(angle_x);
    double cosy = cos(angle_y);
    double cosz = cos(angle_z);
    double sinx = sin(angle_x);
    double siny = sin(angle_y);
    double sinz = sin(angle_z);

    matrix[0][0] =  cosy * cosz;
    matrix[0][1] =  sinx * siny * cosz + cosx * sinz;
    matrix[0][2] = -cosx * siny * cosz + sinx * sinz;
    matrix[1][0] = -cosy * sinz;
    matrix[1][1] = -sinx * siny * sinz + cosx * cosz;
    matrix[1][2] =  cosx * siny * sinz + sinx * cosz;
    matrix[2][0] =  siny;
    matrix[2][1] = -sinx * cosy;
    matrix[2][2] =  cosx * cosy;

}

void rotateZYX(double matrix[3][3], const double angle_x, 
               const double angle_y, const double angle_z)
{
    if (angle_x == 0 && angle_y == 0 && angle_z == 0) 
    {
	for (int j = 0; j < 3; j++)
	    for (int i = 0; i < 3; i++)
		matrix[j][i] = (i == j ? 1 : 0 );
	return;
    }

    /*
      matrix to first rotate reference frame angle_z radians through 
      z axis, then angle_y radians through y axis, and lastly 
      angle_x radians through x axis.  Positive rotations are counter-
      clockwise looking down the axis.
    */
    double cosx = cos(angle_x);
    double cosy = cos(angle_y);
    double cosz = cos(angle_z);
    double sinx = sin(angle_x);
    double siny = sin(angle_y);
    double sinz = sin(angle_z);

    matrix[0][0] =  cosy * cosz;
    matrix[0][1] =  cosy * sinz;
    matrix[0][2] = -siny;
    matrix[1][0] = -cosx * sinz + sinx * siny * cosz;
    matrix[1][1] =  sinx * siny * sinz + cosx * cosz;
    matrix[1][2] =  sinx * cosy;
    matrix[2][0] =  cosx * siny * cosz + sinx * sinz;
    matrix[2][1] = -sinx * cosz + cosx * siny * sinz;
    matrix[2][2] =  cosx * cosy;
}

/*
  Check for the existence of an image file in the current directory,
  the images/ subdirectory in the current directory, the -mapdir
  directory (if specified), and the prefix directory specified in
  auxfiles.h.  
*/
int
fileExists(string &filename)
{
    string newname;

    if (opts.debug) cout << "\nLooking for image file ...\n";

    if (opts.mapdir.size() > 0)
    {
	list<string>::iterator p = opts.mapdir.begin();
	while (p != opts.mapdir.end())
	{
	    newname = *p;
	    newname += separator;
	    newname += filename;
	    
	    if (opts.debug) cout << newname << endl;
	    if (fopen(newname.c_str(), "r") != NULL)
	    {
		filename = newname;
		return(1);
	    }
	    
	    p++;
	}
    }

    if (opts.debug) cout << filename << endl;
    if (fopen(filename.c_str(), "r") != NULL) return(1);

    newname = "images";
    newname += separator;
    newname += filename;

    if (opts.debug) cout << newname << endl;
    if (fopen(newname.c_str(), "r") != NULL)
    {
        filename = newname;
        return(1);
    }

    newname = prefix;
    newname += separator;
    newname += "images";
    newname += separator;
    newname += filename;

    if (opts.debug) cout << newname << endl;
    if (fopen(newname.c_str(), "r") != NULL)
    {
        filename = newname;
        return(1);
    }

    return(0);
}

void
deleteImage(unsigned char *&rgb_data)
{
    if (rgb_data != NULL) delete_image(&rgb_data);
}

//  readImageFile
//
//  Save the RGB pixel array of the image file into rgb_data.  If
//  width or height are < 0, width and height are set to image file's
//  dimensions.  Otherwise rgb_data is scaled to supplied width &
//  height.  
int
readImageFile(string filename, int &width, int &height, 
              unsigned char *&rgb_data)
{
    int image_width, image_height;

    if (!fileExists(filename)) return(0);

    if (!read_image(filename.c_str(), &image_width, &image_height, 
		    &rgb_data))
    {
	return(0);
    }
  
    if (width < 0 || height < 0)
    {
        width = image_width;
        height = image_height;
    }
    else
    {
        if (opts.projection == RECTANGULAR)
            resize_image(&rgb_data, image_width, image_height, 
                         width, height, opts.blend);
        else
            resize_image(&rgb_data, image_width, image_height,
                         width, height, 0);
    }
    
    return(1);
}

void
getColor(const double lon, const double lat, unsigned char pixel[3]) 
{
    if (opts.blend == 0)
    {
        int ilon, ilat;
        image_rect->sphericalToPixel(lon, lat, ilon, ilat);
        memcpy(pixel, image_rect->rgb_data + 
               3 * (ilat*image_rect->getWidth() + ilon), 3);
        return;
    }

    /*
      The point (lon, lat) falls into one of these four pixels.  The 
      color returned will be an area weighted average of these four.
    
      --- ---
      | 0 | 1 |
      --- ---
      | 2 | 3 |
      --- ---
    */
  
    int ix[4], iy[4];

    image_rect->sphericalToPixel(lon, lat, ix[1], iy[2]);

    if (ix[1] == 0)
    {
        ix[0] = image_rect->getWidth() - 1;
        ix[1] = 0;
    }
    else
    {
        ix[0] = ix[1] - 1;
    }

    iy[0] = iy[2] - 1;
    if (iy[0] == -1) iy[0] = 0;

    ix[2] = ix[0];
    ix[3] = ix[1];
    iy[1] = iy[0];
    iy[3] = iy[2];
  
    double t = lon - image_rect->xToLon(ix[2]);
    if (t > M_PI) t -= TWO_PI;
    if (t < -M_PI) t += TWO_PI;
    t /= image_rect->getDeltaLon();

    double u = ((lat - image_rect->yToLat(iy[2]))
                / image_rect->getDeltaLat());

    // Weights are from Numerical Recipes, 2nd Edition
    //        weight[0] = (1 - t) * u;
    //        weight[2] = (1-t) * (1-u);
    //        weight[3] = t * (1-u);
    double weight[4];
    weight[1] = t * u;
    weight[0] = u - weight[1];
    weight[2] = 1 - t - u + weight[1];
    weight[3] = t - weight[1];

    memset(pixel, 0, 3);
    for (int ii = 0; ii < 4; ii++)
    {
        int ipos = 3*(iy[ii]*image_rect->getWidth() + ix[ii]);
	pixel[0] += (unsigned char) 
	    (weight[ii] * image_rect->rgb_data[ipos++]);
	pixel[1] += (unsigned char) 
	    (weight[ii] * image_rect->rgb_data[ipos++]);
	pixel[2] += (unsigned char) 
	    (weight[ii] * image_rect->rgb_data[ipos++]);
    }
}

/* Based on Chapter 18 of Astronomical Formulae for Calculators by Meeus */
void 
sunpos(const double T)
{
    double L = (poly(2.7969668E2, 3.600076892E4, 3.025E-4, 0, T) 
                * deg_to_rad);
    double M = (poly(3.5847583E2, 3.599904975E4, -1.5E-4, -3.3E-6, T) 
                * deg_to_rad);
    double ecc = poly(1.675104E-2, -4.18E-5, -1.26E-7, 0, T);
    double eccanom = kepler(ecc, M);
    double nu = 2 * atan(sqrt((1 + ecc) / (1 - ecc)) * tan(eccanom / 2));
    double theta = L + nu - M;

    sunlon = theta;
    sunlat = 0;
    sundist = 1.0000002 * (1 - ecc * cos(eccanom));
}

double
getSunLat()
{
    return(sunlat);
}

double
getSunLon()
{
    return(sunlon);
}

double
getSunDist()
{
    return(sundist);
}

void
setSunLat(double lat)
{
    sunlat = lat;
}

void
setSunLon(double lon)
{
    sunlon = lon;
}

double
getMoonLat()
{
    return(moonlat);
}

double
getMoonLon()
{
    return(moonlon);
}

double
getMoonDist()
{
    return(moondist);
}

void
setMoonLat(double lat)
{
    moonlat = lat;
}

void
setMoonLon(double lon)
{
    moonlon = lon;
}

